/*

  Copyright (C) 1999 President and Fellows of Harvard College

  All rights reserved.

  NOTICE: This software is provided "as is", without any warranty,
  including any implied warranty for merchantability or fitness for a
  particular purpose.  Under no circumstances shall Harvard University
  or its faculty, staff, students or agents be liable for any use of,
  misuse of, or inability to use this software, including incidental
  and consequential damages.

  License is hereby given to use, modify, and redistribute this
  software, in whole or in part, for any commercial or non-commercial
  purpose, provided that the user agrees to the terms of this
  copyright notice, including disclaimer of warranty, and provided
  that this copyright notice, including disclaimer of warranty, is
  preserved in the source code and documentation of anything derived
  from this software.  Any redistributor of this software or anything
  derived from this software assumes responsibility for ensuring that
  any parties to whom wuch a redistribution is made are fully aware of
  the terms of this license and disclaimer.

  Ported from Linux and extended by:
    Brad Karp, Harvard University EECS, April, 1999
    karp@eecs.harvard.edu

  The above copyright and licensing notice applies to the
  *modifications* made at Harvard to port and extend the Linux driver,
  whose original GPL license appears below. Note that the GPL license
  is also in force on this FreeBSD port of the driver.

*/

/********************************************************************
 *	WaveLAN/IEEE PCMCIA Linux network device driver
 *
 *	by Andreas Neuhaus <andy@fasta.fh-dortmund.de>
 *	http://www.fasta.fh-dortmund.de/users/andy/wvlan/
 *
 *	This driver is free software; you can redistribute and/or
 *	modify it under the terms of the GNU General Public License;
 *	either version 2 of the license, or (at your option) any
 *	later version.
 *	Please send comments/bugfixes/patches to the above address.
 *
 *	Based on the Linux PCMCIA Programmer's Guide
 *	and Lucent Technologies' HCF-Light library (wvlan47).
 *
 *	See wvlan.README for more information.
 */

#include "iwl.h"
#include "bpfilter.h"

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/sockio.h>
#include <sys/malloc.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/syslog.h>
#include <sys/proc.h>
#include <sys/ucred.h>

#include <i386/isa/if_iwl_osindep.h>

#include <net/ethernet.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <net/if_dl.h>
#include <net/if_mib.h>

#if NBPFILTER > 0
#include <net/bpf.h>
#endif

#include <machine/clock.h>
#include <machine/md_var.h>

#include <i386/isa/isa_device.h>
#include <i386/isa/icu.h>

#include "card.h"

struct iwl_softc {
  struct arpcom arpcom;		/* ethernet common */
  u_char	gone;		/* HW missing, presumed having a good time */
  IFB_STRCT ifb;		/* os-indep interface status block */
  struct	ifmib_iso_8802_3 mibdata; /* stuff for network mgmt */
};

static struct iwl_softc iwl_softc[NIWL];

/* Module parameters */
static int port_type = 0x01;		/* default 1=BSS */
static char *station_name = "\0";	/* default card-configured */
static char *network_name = "\0";	/* default card-configured */
static int channel = 3;			/* default 1st channel */

static void iwl_watchdog	__P((struct ifnet *));

/* Valid MTU values (2304 is the max hardware frame size) */
#define IWL_MIN_MTU 256
#define IWL_MAX_MTU 2304-8

/* macros for cookie-cutter routines to query a short-valued parameter from
   NIC */
#ifdef IWL_DEBUG
#define DEB_HWG_SIMPLE printf("iwl: %s:0x%x returned 0x%x\n", dstr, code, ltv.id[0])
#else
#define DEB_HWG_SIMPLE
#endif

#define IWL_HWG_SIMPLE(code, dstr) \
  int rc;                               \
  CFG_ID_STRCT ltv;                     \
  ltv.len = 2;                          \
  ltv.typ = (code);                     \
  rc = hcf_get_info(ifbp, (LTVP) &ltv); \
  DEB_HWG_SIMPLE;                       \
  return rc ? -1 : ltv.id[0]

/* Ethernet-II snap header */
static char snap_header[] = { 0x00, 0x00, 0xaa, 0xaa, 0x03, 0x00, 0x00, 0xf8 };

int iwl_hw_getmacaddr __P((IFBP, u_char *, int));
int iwl_hw_getchannellist __P((IFBP));
int iwl_hw_getporttype __P((IFBP));
int iwl_hw_setporttype __P((IFBP, int));
int iwl_hw_getstationname __P((IFBP, char *, int));
int iwl_hw_setstationname __P((IFBP, char *));
int iwl_hw_getssid __P((IFBP, char *, int));
int iwl_hw_setssid __P((IFBP, char *));
int iwl_hw_getchannel __P((IFBP));
int iwl_hw_setchannel __P((IFBP, int));
int iwl_hw_getcurrentchannel __P((IFBP));
int iwl_hw_getpower __P((IFBP));
int iwl_hw_setpower __P((IFBP, int));
int iwl_hw_getrtxs __P((IFBP));
int iwl_hw_getrtxl __P((IFBP));

#if 0
/* karp: these appear only to be for monitoring: port later. */
int wvlan_hw_getcurrentssid (IFBP ifbp, char *name, int len);
#endif

int iwl_hw_config __P((struct isa_device *));
int iwl_hw_shutdown __P((struct iwl_softc *));
int iwl_hw_reset __P((struct ifnet *));

static void iwl_setmulti __P((struct ifnet *));
static int iwl_ioctl __P((register struct ifnet *, u_long, caddr_t));
#ifdef 0
struct iw_statistics *wvlan_get_wireless_stats (struct device *dev);
#endif

static void iwl_init __P((void *));
static int iwl_attach __P((struct iwl_softc *, int, int));
static void iwl_start __P((struct ifnet *));
void iwlintr_sc __P((struct iwl_softc *));
static __inline void iwl_rint __P((struct iwl_softc *, int));

static int iwl_probe_dummy __P((struct isa_device *));
static int iwl_attach_dummy __P((struct isa_device *));

static void iwl_init (void *xsc)
{
  struct iwl_softc *sc = xsc;
  struct ifnet *ifp = &sc->arpcom.ac_if;
  int rc, i, chlist, s;

  if (sc->gone)
    return;

  /* address not known */
  if (TAILQ_EMPTY(&ifp->if_addrhead)) /* unlikely? XXX */
    return;

  s = splimp();

  ifp->if_timer = 0;

  rc = hcf_disable(&sc->ifb, 0);
#ifdef IWL_DEBUG
  printf("iwl%d: hcf_disable(0) returned 0x%x\n", ifp->if_unit, rc);
#endif
  /* Set hardware parameters */
  rc = iwl_hw_setporttype(&sc->ifb, port_type);
  if (rc)
#ifdef IWL_DEBUG
    printf("iwl%d: iwl_hw_setporttype(%d) returned 0x%x\n", ifp->if_unit,
	   port_type, rc);
#endif
  if (*station_name)
    rc = iwl_hw_setstationname(&sc->ifb, station_name);
  if (rc)
#ifdef IWL_DEBUG
    printf("iwl%d: iwl_hw_setstationname(%s) returned 0x%x\n", ifp->if_unit,
	   station_name, rc);
#endif
  if (*network_name)
    rc = iwl_hw_setssid(&sc->ifb, network_name);
  if (rc)
#ifdef IWL_DEBUG
    printf("iwl%d: iwl_hw_setssid(%s) returned 0x%x\n", ifp->if_unit,
	   network_name, rc);
#endif

  /* Check valid channel settings */
  if (port_type == 3) {
      chlist = iwl_hw_getchannellist(&sc->ifb);
      printf("iwl%d: Valid channels: ", ifp->if_unit);
      for (i=1; i<17; i++)
	if (1<<(i-1) & chlist)
	  printf("%d ", i);
      printf("\n");
      if (channel<1 || channel>16 || !(1<<(channel-1) & chlist))
	printf("iwl%d: Channel value of %d is invalid!\n", ifp->if_unit,
	       channel);
      else
	rc = iwl_hw_setchannel(&sc->ifb, channel);
#ifdef IWL_DEBUG
      if (rc)
	printf("iwl%d: iwl_hw_setchannel(%d) returned 0x%x\n", ifp->if_unit,
	       channel, rc);
#endif
  }

  /* interrupts on */
  rc = hcf_action(&sc->ifb, HCF_ACT_INT_ON);
#ifdef IWL_DEBUG
  printf("iwl%d: hcf_action(HCF_ACT_INT_ON) returned 0x%x\n",
	 ifp->if_unit, rc);
#endif

  /* Enable hardware */
  rc = hcf_enable(&sc->ifb, 0);
#ifdef IWL_DEBUG
  printf("iwl: hcf_enable(0) returned 0x%x\n", rc);
#endif

  ifp->if_flags |= IFF_RUNNING;
  ifp->if_flags &= ~IFF_OACTIVE;

#if 0
  /* karp: doesn't seem to make sense for WaveLAN: the card will not let us
     ask to start sending unless we've pio'ed a packet to it, which shouldn't
     have been possible before here */
  /* start output on the interface */
  iwl_start(ifp);
#endif
  (void) splx(s);
}

static void
iwl_watchdog(struct ifnet *ifp)
{
  struct iwl_softc *sc = ifp->if_softc;

  if (sc->gone)
    return;
  log(LOG_ERR, "iwl%d: device timeout\n", ifp->if_unit);
  ifp->if_oerrors++;

  ifp->if_flags &= ~IFF_OACTIVE;
  iwl_hw_reset(ifp);
}

/*
 * Install interface into kernel networking data structures
 */
static int
iwl_attach(struct iwl_softc *sc, int unit, int flags)
{
	struct ifnet *ifp = &sc->arpcom.ac_if;
	int rc;

	/*
	 * Set interface to stopped condition (reset)
	 */
	hcf_disable(&sc->ifb, 0);

	if (!ifp->if_name) {
		/*
		 * Initialize ifnet structure
		 */
		ifp->if_softc = sc;
		ifp->if_unit = unit;
		ifp->if_name = "iwl";
		ifp->if_output = ether_output;
		ifp->if_start = iwl_start;
		ifp->if_ioctl = iwl_ioctl;
		ifp->if_watchdog = iwl_watchdog;
		ifp->if_init = iwl_init;
		ifp->if_snd.ifq_maxlen = IFQ_MAXLEN;
		ifp->if_linkmib = &sc->mibdata;
		ifp->if_linkmiblen = sizeof sc->mibdata;

		/* XXX someday, set mibdata.dot3StatsEtherChipSet and
		   mibdata.dot3Compliance here */

		ifp->if_flags = (IFF_BROADCAST | IFF_SIMPLEX |
				 IFF_MULTICAST);

		/*
		 * Attach the interface
		 */
		if_attach(ifp);
		ether_ifattach(ifp);
	}

	/*
	 * Print additional info when attached
	 */
	printf("%s%d: address %6D\n", ifp->if_name, ifp->if_unit, 
		sc->arpcom.ac_enaddr, ":");

	/*
	 * If BPF is in the kernel, call the attach for it
	 */
#if NBPFILTER > 0
	bpfattach(ifp, DLT_EN10MB, sizeof(struct ether_header));
#endif
	return 1;
}

static __inline void
iwl_rint(struct iwl_softc *sc, int len)
{
  struct ether_header *eh;
  struct ifnet *ifp = &sc->arpcom.ac_if;
  struct mbuf *m;

  if (sc->gone)
    return;

  /* grab mbuf */
  MGETHDR(m, M_DONTWAIT, MT_DATA);
  if (!m)
    return;

  m->m_pkthdr.rcvif = &sc->arpcom.ac_if;
  m->m_pkthdr.len = m->m_len = len;

  if ((len + 2) > MHLEN) {
    MCLGET(m, M_DONTWAIT);
    if ((m->m_flags & M_EXT) == 0) {
      m_freem(m);
      return;
    }
  }
  /* round the start of the IP address to a 4-byte boundary */
  m->m_data += 2;
  eh = mtod(m, struct ether_header *);

#ifdef IWL_DEBUG
  printf("iwl%d: Receiving 0x%x octets\n", ifp->if_unit, len);
#endif

  /* Add Ethernet-II frame decapsulation, because
     HCF-light doesn't support that. */
  if (sc->ifb.IFB_RxStat == 0x2000 || sc->ifb.IFB_RxStat == 0x4000) {
    hcf_get_data(&sc->ifb, 0, (caddr_t) eh, 12);
    hcf_get_data(&sc->ifb, 12+sizeof(snap_header), ((caddr_t) eh)+12,
		 len-12-sizeof(snap_header));
    m->m_pkthdr.len = m->m_len = len - sizeof(snap_header);
  }
  else
    hcf_get_data(&sc->ifb, 0, (caddr_t) eh, len);

#if NBPFILTER > 0

  if (sc->arpcom.ac_if.if_bpf) {
    bpf_mtap(&sc->arpcom.ac_if, m);

    /* if we're PROMISC, which happens both for ALLMULTI and PROMISC, *and*
       when we've joined too many multicast groups to use the h/w filter,
       then only let the packet through to the ether layer if it's
       multicast, or if it's unicast with our ether address */
    if (((sc->arpcom.ac_if.if_flags & (IFF_PROMISC | IFF_ALLMULTI)) || 
	 (sc->arpcom.ac_multicnt > 16)) &&
	(eh->ether_dhost[0] & 1) == 0 &&
	bcmp(eh->ether_dhost, sc->arpcom.ac_enaddr,
	     sizeof(eh->ether_dhost)) != 0 &&
	bcmp(eh->ether_dhost, etherbroadcastaddr,
	     sizeof(eh->ether_dhost)) != 0) {
      m_freem(m);
      return;
    }
  }
#endif

  m->m_pkthdr.len = m->m_len = len - sizeof(struct ether_header);
  m->m_data += sizeof(struct ether_header);

  ether_input(&sc->arpcom.ac_if, eh, m);
}

void
iwlintr_sc(struct iwl_softc *sc)
{
  struct ifnet *ifp = &sc->arpcom.ac_if;
  int cnt = 5;
  int ev, s, len;

  if (sc->gone)
    return;

  s = splbio();

  /* Process pending interrupts.
     We continue until hcf_service_nic tells that no received
     frames are pending. However we should check to not lock up
     here in an endless loop. */
  /* karp: is there really a problem with the NIC delivering an endless stream
     of interrupts??? */
  while (cnt--) {
    /* Ask NIC why interrupt occured */
    ev = hcf_service_nic(&sc->ifb);
#ifdef IWL_DEBUG
    printf("iwl%d: hcf_service_nic() returned 0x%x RscInd 0x%x if_flags %ld\n", ifp->if_unit, ev, sc->ifb.IFB_PIFRscInd, (long) ifp->if_flags & IFF_OACTIVE);
#endif
    /* Transmission completion seem to be also signalled with ev==0
       better check that out with RscInd and complete transfer also */
    if (sc->ifb.IFB_PIFRscInd && (ifp->if_flags & IFF_OACTIVE))
      ev |= HREG_EV_TX;

    /* HREG_EV_RES: WMAC controller H/W error (wait timeout) */
    if (ev & HREG_EV_RES) {
      /* This message seems to occur often on heavy load
	 but it seems not to affect transmission,
	 so we simply ignore it. */
#ifdef IWL_DEBUG
      printf("iwl%d: WMAC controller hw error (wait timeout)\n", ifp->if_unit);
#endif
    }

    /* HREG_EV_INFO_DROP: WMAC did not have sufficient RAM to build
       unsolicited frame */
    if (ev & HREG_EV_INFO_DROP)
      printf("iwl%d: WMAC did not have sufficient RAM to build unsolicited frame!\n", ifp->if_unit);

    /* HREG_EV_INFO: WMAC controller asynchronous information frame */
    if (ev & HREG_EV_INFO) {
#ifdef IWL_DEBUG
      printf("iwl%d: WMAC controller asynchronous information frame\n",
	     ifp->if_unit);
#endif
    }

    /* HREG_EV_TX_EXC: WMAC controller asynchronous transmission unsuccessfully
       completed */
    if (ev & HREG_EV_TX_EXC) {
      printf("iwl%d: WMAC controller transmit failure\n", ifp->if_unit);
      ifp->if_flags &= ~IFF_OACTIVE;
      ifp->if_oerrors++;
      ifp->if_timer = 0;
      /* XXX karp: someday, update MIB data here */
    }

    /* HREG_EV_TX: WMAC controller asynchronous transmission successfully
       completed */
    if (ev & HREG_EV_TX) {
      if (!(ifp->if_flags & IFF_OACTIVE))
	printf("iwl%d: transmit complete, but no transmit active!\n",
	       ifp->if_unit);
#ifdef IWL_DEBUG
      printf("iwl%d: packet transmit completed\n", ifp->if_unit);
#endif
      ifp->if_flags &= ~IFF_OACTIVE;
      ifp->if_opackets++;
      ifp->if_timer = 0;
    }

    /* HREG_EV_RX: WMAC controller asynchronous receive frame */
    /* Break loop if no frame was received. */
    if (!(ev & HREG_EV_RX))
      break;

    /* If a frame was received, we process it and wrap back
       up to the top of the while(1) loop so that hcf_service_nic()
       gets called again after the frame drained from the NIC.
       This allows us to find out if yet another frame has
       arrived, and also to immediately acknowledge the just-
       processed frame so that the NIC's buffer gets de-
       allocated right away. */
    len = sc->ifb.IFB_RxLen;
    if (len) {
#ifdef IWL_DEBUG
      printf("iwl%d: Frame received. rx_len=0x%x\n", ifp->if_unit, len);
#endif
      iwl_rint(sc, len);
      ifp->if_ipackets++;
    }
  }

  if (!cnt)
    printf("iwl%d: too many interrupt loops!\n", ifp->if_unit);

  (void) splx(s);

  /* gave receive priority. now drain a packet from ifq if the output side of
     the interface is idle. */
  if (!(ifp->if_flags & IFF_OACTIVE))
    iwl_start(ifp);
}

#if NCARD > 0
#include <sys/kernel.h>
#include <sys/select.h>
#include <sys/module.h>
#include <pccard/cardinfo.h>
#include <pccard/slot.h>

/*
 *	PC-Card (PCMCIA) specific code.
 */
static int	iwlinit		__P((struct pccard_devinfo *));
static void	iwlunload	__P((struct pccard_devinfo *));
static int	card_intr	__P((struct pccard_devinfo *));

PCCARD_MODULE(iwl, iwlinit, iwlunload, card_intr, 0, net_imask);

/*
 *	Initialize the device - called from Slot manager.
 */
static int
iwlinit(struct pccard_devinfo *devi)
{
	struct iwl_softc *sc = &iwl_softc[devi->isahd.id_unit];

#ifdef IWL_DEBUG
	printf("iwl%d: trying port 0x%x\n", devi->isahd.id_unit,
	       devi->isahd.id_iobase);
#endif
	/* validate unit number. */
	if (devi->isahd.id_unit >= NIWL)
		return(ENODEV);
	/*
	 * Probe the device. If a value is returned, the
	 * device was found at the location.
	 */
	sc->gone = 0;
	/* note: we ignore the misc parameter of devi, because pccardd
	   doesn't know how to find an Ethernet address for a WaveLAN card */
	if (iwl_hw_config(&devi->isahd) != 0)
	  return(ENXIO);
	if (iwl_attach(sc, devi->isahd.id_unit, devi->isahd.id_flags) == 0)
		return(ENXIO);

	return(0);
}

/*
 *	iwlunload - unload the driver and clear the table.
 */
static void
iwlunload(struct pccard_devinfo *devi)
{
  struct iwl_softc *sc = &iwl_softc[devi->isahd.id_unit];
  struct ifnet *ifp = &sc->arpcom.ac_if;

  if (sc->gone) {
    printf("iwl%d: already unloaded\n", devi->isahd.id_unit);
    return;
  }
  ifp->if_flags &= ~IFF_RUNNING;
  if_down(ifp);
  iwl_hw_shutdown(sc);
  sc->gone = 1;
  printf("iwl%d: unload\n", devi->isahd.id_unit);
}

/*
 *	card_intr - Shared interrupt called from
 *	front end of PC-Card handler.
 */
static int
card_intr(struct pccard_devinfo *devi)
{
	iwlintr_sc(&iwl_softc[devi->isahd.id_unit]);
	return(1);
}
#endif /* NCARD > 0 */

/* this driver is only for pccards; ISA probe always fails */
static int iwl_probe_dummy(struct isa_device *dummy)
{
  return 0;
}

/* should never be called; fake failure if it ever is */
static int iwl_attach_dummy(struct isa_device *dummy)
{
  return 0;
}

struct isa_driver iwldriver = {
	iwl_probe_dummy,
	iwl_attach_dummy,
	"iwl"
};

/********************************************************************
 * HARDWARE SETTINGS
 */
int iwl_hw_getmacaddr (IFBP ifbp, u_char *mac, int len)
{
  CFG_MAC_ADDR_STRCT ltv;
  int rc, l;

  ltv.len = 4;
  ltv.typ = CFG_CNF_OWN_MAC_ADDR;
  rc = hcf_get_info(ifbp, (LTVP) &ltv);
#ifdef IWL_DEBUG
  printf("iwl: hcf_get_info(CFG_CNF_OWN_MAC_ADDR) returned 0x%x\n", rc);
#endif
  if (rc)
    return rc;
  l = min(len, ltv.len*2);
  bcopy((u_char *)ltv.mac_addr, mac, l);
  return 0;
}

int iwl_hw_getchannellist (IFBP ifbp)
{
  CFG_ID_STRCT ltv;
  int rc, chlist;

  ltv.len = 2;
  ltv.typ = CFG_CHANNEL_LIST;
  rc = hcf_get_info(ifbp, (LTVP) &ltv);
  chlist = ltv.id[0];
#ifdef IWL_DEBUG
  printf("iwl: hcf_get_info(CFG_CHANNEL_LIST):0x%x returned 0x%x\n",
	 chlist, rc);
#endif
  return rc ? 0 : chlist;
}

int iwl_hw_setporttype (IFBP ifbp, int ptype)
{
  CFG_ID_STRCT ltv;
  int rc;

  ltv.len = 2;
  ltv.typ = CFG_CNF_PORT_TYPE;
  ltv.id[0] = ptype;
  rc = hcf_put_info(ifbp, (LTVP) &ltv);
#ifdef IWL_DEBUG
  printf("iwl: hcf_put_info(CFG_CNF_PORT_TYPE:0x%x) returned 0x%x\n",
	 ptype, rc);
#endif
  return rc;
}

int iwl_hw_getporttype (IFBP ifbp)
{
  IWL_HWG_SIMPLE(CFG_CNF_PORT_TYPE, "CFG_CNF_PORT_TYPE");
}

int iwl_hw_setstationname (IFBP ifbp, char *name)
{
  CFG_ID_STRCT ltv;
  int rc, l;

  ltv.len = 18;
  ltv.typ = CFG_CNF_OWN_NAME;
  l = min(strlen(name), (ltv.len-2)*2);
  ltv.id[0] = l;
  bcopy(name, (char *) &ltv.id[1], l);
  rc = hcf_put_info(ifbp, (LTVP) &ltv);
#ifdef IWL_DEBUG
  printf("iwl: hcf_put_info(CFG_CNF_OWN_NAME:'%s') returned 0x%x\n", name, rc);
#endif
  return rc;
}

int iwl_hw_getstationname (IFBP ifbp, char *name, int len)
{
  CFG_ID_STRCT ltv;
  int rc, l;

  ltv.len = 18;
  ltv.typ = CFG_CNF_OWN_NAME;
  rc = hcf_get_info(ifbp, (LTVP) &ltv);
#ifdef IWL_DEBUG
  printf("iwl: hcf_get_info(CFG_CNF_OWN_NAME) returned 0x%x\n", rc);
#endif
  if (rc)
    return rc;
  if (ltv.id[0]) {
    l = min(len, ltv.id[0]);
    memcpy(name, (char *) &ltv.id[1], l);
  }
  else {
    l = strlen("<none>");
    memcpy(name, "<none>", l);
  }
  if (l == len)
    l--;
  name[l] = 0;
#ifdef IWL_DEBUG
  printf("iwl: hcf_get_info(CFG_CNF_OWN_NAME):'%s'\n", name);
#endif
  return 0;
}

int iwl_hw_setssid (IFBP ifbp, char *name)
{
  CFG_ID_STRCT ltv;
  int rc, l;

  ltv.len = 18;
  if (port_type == 3)
    ltv.typ = CFG_CNF_OWN_SSID;
  else
    ltv.typ = CFG_CNF_DESIRED_SSID;
  l = min(strlen(name), (ltv.len-2)*2);
  ltv.id[0] = l;
  bcopy(name, (char *) &ltv.id[1], l);
  rc = hcf_put_info(ifbp, (LTVP) &ltv);
#ifdef IWL_DEBUG
  printf("iwl: hcf_put_info(CFG_CNF_OWN/DESIRED_SSID:'%s') returned 0x%x\n",
	 name, rc);
#endif
  return rc;
}

int iwl_hw_getssid (IFBP ifbp, char *name, int len)
{
  CFG_ID_STRCT ltv;
  int rc, l;

  ltv.len = 18;
  if (port_type == 3)
    ltv.typ = CFG_CNF_OWN_SSID;
  else
    ltv.typ = CFG_CNF_DESIRED_SSID;
  rc = hcf_get_info(ifbp, (LTVP) &ltv);
#ifdef IWL_DEBUG
  printf("iwl: hcf_get_info(CFG_CNF_OWN/DESIRED_SSID) returned 0x%x\n", rc);
#endif
  if (rc)
    return rc;
  if (ltv.id[0]) {
    l = min(len, ltv.id[0]);
    memcpy(name, (char *) &ltv.id[1], l);
  }
  else {
    l = strlen("<any>");
    memcpy(name, "<any>", l);
  }
  if (l == len)
    l--;
  name[l] = 0;
#ifdef IWL_DEBUG
  printf("iwl: hcf_get_info(CFG_CNF_OWN/DESIRED_SSID):'%s'\n", name);
#endif
  return 0;
}

#if 0
/* karp: port later */
int wvlan_hw_getcurrentssid (IFBP ifbp, char *name, int len)
{
  CFG_ID_STRCT ltv;
  int rc, l;

  ltv.len = 18;
  if (port_type == 3)
    ltv.typ = CFG_CURRENT_BSSID;
  else
    ltv.typ = CFG_CURRENT_SSID;
  rc = hcf_get_info(ifbp, (LTVP) &ltv);
  DEBUG(DEBUG_NOISY, "%s: hcf_get_info(CFG_CURRENT_(B)SSID) returned 0x%x\n", dev_info, rc);
  if (rc)
    return rc;
  if (ltv.id[0])
    l = min(len, ltv.id[0]);
  else
    l = min(len, ltv.len*2);
  memcpy(name, (char *) &ltv.id[1], l);
  name[l] = 0;
  DEBUG(DEBUG_NOISY, "%s: hcf_get_info(CFG_CURRENT_(B)SSID):'%s'\n", dev_info, name);
  return 0;
}
#endif

int iwl_hw_setchannel (IFBP ifbp, int channel)
{
  CFG_ID_STRCT ltv;
  int rc;

  ltv.len = 2;
  ltv.typ = CFG_CNF_OWN_CHANNEL;
  ltv.id[0] = channel;
  rc = hcf_put_info(ifbp, (LTVP) &ltv);
#ifdef IWL_DEBUG
  printf("iwl: hcf_put_info(CFG_CNF_OWN_CHANNEL:0x%x) returned 0x%x\n",
	 channel, rc);
#endif
  return rc;
}

int iwl_hw_getchannel (IFBP ifbp)
{
  IWL_HWG_SIMPLE(CFG_CNF_OWN_CHANNEL, "CFG_CNF_OWN_CHANNEL");
}

int iwl_hw_getcurrentchannel (IFBP ifbp)
{
  IWL_HWG_SIMPLE(CFG_CURRENT_CHANNEL, "CFG_CURRENT_CHANNEL");
}

int iwl_hw_getpower(IFBP ifbp)
{
  IWL_HWG_SIMPLE(CFG_CNF_PM_ENABLED, "CFG_CNF_PM_ENABLED");
}

int iwl_hw_setpower(IFBP ifbp, int pm)
{
  CFG_ID_STRCT ltv;
  int rc;

  ltv.len = 2;
  ltv.typ = CFG_CNF_PM_ENABLED;
  ltv.id[0] = pm;
  rc = hcf_put_info(ifbp, (LTVP) &ltv);
#ifdef IWL_DEBUG
  printf("iwl: hcf_put_info(CFG_CNF_PM_ENABLED):0x%x returned 0x%x\n",
	 pm, rc);
#endif
  return rc;
}

int iwl_hw_getrtxs(IFBP ifbp)
{
  IWL_HWG_SIMPLE(CFG_SHORT_RETRY_LIMIT, "CFG_SHORT_RETRY_LIMIT");
}

int iwl_hw_getrtxl(IFBP ifbp)
{
  IWL_HWG_SIMPLE(CFG_LONG_RETRY_LIMIT, "CFG_LONG_RETRY_LIMIT");
}

/********************************************************************
 * HARDWARE CONFIG / SHUTDOWN / RESET
 */
int iwl_hw_config (struct isa_device *isahd)
{
  struct iwl_softc *sc = &iwl_softc[isahd->id_unit];
  int rc;

#ifdef IWL_DEBUG
  printf("-> iwl_hw_config(%d)\n", isahd->id_unit);
#endif

  /* Init the HCF library */
  hcf_connect(&sc->ifb, isahd->id_iobase);

  /* Init hardware */
  rc = hcf_action(&sc->ifb, HCF_ACT_CARD_IN);
#ifdef IWL_DEBUG
  printf("iwl%d: hcf_action(HCF_ACT_CARD_IN) returned 0x%x\n",
	 isahd->id_unit, rc);
#endif
#if 0 && defined(PCMCIA_DEBUG) && (PCMCIA_DEBUG>=DEBUG_INTERRUPT) /* Linux */
  local->ifb.IFB_IntEnMask |= HREG_EV_TICK;
#endif
  /* Get MAC address */
  rc = iwl_hw_getmacaddr(&sc->ifb, sc->arpcom.ac_enaddr, ETHER_ADDR_LEN);
  if (rc)
    return rc;

#ifdef IWL_DEBUG
  printf("<- iwl_hw_config()\n");
#endif
  return 0;
}

int iwl_hw_shutdown (struct iwl_softc *sc)
{
  int rc;

#ifdef IWL_DEBUG
  printf("-> iwl_hw_shutdown()\n");
#endif

  /* Disable and shutdown hardware */
  rc = hcf_action(&sc->ifb, HCF_ACT_CARD_OUT);
#ifdef IWL_DEBUG
  printf("iwl: hcf_action(HCF_ACT_CARD_OUT) returned 0x%x\n", rc);
#endif

  /* Release HCF library */
  hcf_disconnect(&sc->ifb);

#ifdef IWL_DEBUG
  printf("<- iwl_hw_shutdown()\n");
#endif
  return 0;
}

int iwl_hw_reset (struct ifnet *ifp)
{
  struct iwl_softc *sc = ifp->if_softc;
  int rc, s;

#ifdef IWL_DEBUG
  printf("-> iwl_hw_reset(iwl%d)\n", ifp->if_unit);
#endif

  if (sc->gone)
    return 0;

  s = splimp();
  /* Disable hardware */
  rc = hcf_disable(&sc->ifb, 0);
#ifdef IWL_DEBUG
  printf("iwl%d: hcf_disable(0) returned 0x%x\n", ifp->if_unit, rc);
#endif
  rc = hcf_action(&sc->ifb, HCF_ACT_INT_OFF);
#ifdef IWL_DEBUG
  printf("iwl%d: hcf_action(HCF_ACT_INT_OFF) returned 0x%x\n", ifp->if_unit,
	 rc);
#endif

  /* Re-Enable hardware */
  rc = hcf_action(&sc->ifb, HCF_ACT_INT_ON);
#ifdef IWL_DEBUG
  printf("iwl%d: hcf_action(HCF_ACT_INT_ON) returned 0x%x\n", ifp->if_unit,
	 rc);
#endif
  rc = hcf_enable(&sc->ifb, 0);
#ifdef IWL_DEBUG
  printf("iwl%d: hcf_enable(0) returned 0x%x\n", ifp->if_unit, rc);
#endif

#ifdef IWL_DEBUG
  printf("<- iwl_hw_reset()\n");
#endif
  (void) splx(s);
  return rc;
}

struct iwl_groups {
  uint16_t len;
  uint16_t typ;
  uint8_t addrs[96];
};

static void
iwl_setmulti(struct ifnet *ifp)
{
  struct iwl_softc *sc = ifp->if_softc;
  LTV_STRCT ltv;

  ltv.len = 2;
  ltv.typ = CFG_PROMISCUOUS_MODE;
  /* emulate ALLMULTI with PROMISC. */
  if (ifp->if_flags & (IFF_PROMISC | IFF_ALLMULTI)) {
    /* put us in promiscuous mode */
    ltv.val[0] = 1;
    hcf_put_info(&sc->ifb, (LTVP) &ltv);
  }
  else {
    /* out of promiscuous mode, if few enough groups to use hardware */
    ltv.val[0] = (sc->arpcom.ac_multicnt > 16) ? 1 : 0;
    hcf_put_info(&sc->ifb, (LTVP) &ltv);
    /* set multicast filter */
    if (!ltv.val[0]) {
      struct ifmultiaddr *ifma;
      struct iwl_groups ig;
      int i;

      for (i = 0, ifma = sc->arpcom.ac_if.if_multiaddrs.lh_first; ifma;
	   ifma = ifma->ifma_link.le_next) {
	if (ifma->ifma_addr->sa_family != AF_LINK)
	  continue;
	bcopy(LLADDR((struct sockaddr_dl *)ifma->ifma_addr),
	      ((caddr_t)(&ig.addrs))+i*ETHER_ADDR_LEN, ETHER_ADDR_LEN);
	i++;
      }
      ig.len = i*3+1;
      ig.typ = CFG_GROUP_ADDR;
      hcf_put_info(&sc->ifb, (LTVP) &ig);
    }
  }
}

static int
iwl_ioctl(register struct ifnet *ifp, u_long command, caddr_t data)
{
  struct iwl_softc *sc = ifp->if_softc;
  int s, i, error = 0;
  struct ifreq *ifr;
  struct proc *p = curproc;	/* XXX -- is this disallowed?  */
  char tnam[33];
  size_t done;

#ifdef IWL_DEBUG
  printf("-> iwl_ioctl(%d, cmd=0x%x)\n", ifp->if_unit, (unsigned int) command);
#endif

  if (sc->gone) {
    ifp->if_flags &= ~IFF_RUNNING;
    return ENXIO;
  }

  s = splimp();

  switch (command) {

  case SIOCSIFADDR:
  case SIOCGIFADDR:
    error = ether_ioctl(ifp, command, data);
    break;

  case SIOCSIFMTU:
    /* need to do this locally, and not in ether_ioctl(), because 802.11
       allows a larger MTU than standard 802.3 Ethernet */
    ifr = (struct ifreq *) data;
    if (ifr->ifr_mtu > IWL_MAX_MTU || ifr->ifr_mtu < IWL_MIN_MTU)
      error = EINVAL;
    else
      ifp->if_mtu = ifr->ifr_mtu;
    break;

  case SIOCSIFFLAGS:
    /*
     * If the interface is marked up and stopped, then start it.
     * If it is marked down and running, then stop it.
     */
    if (ifp->if_flags & IFF_UP) {
      if ((ifp->if_flags & IFF_RUNNING) == 0)
	iwl_init(sc);
    }
    else {
      if (ifp->if_flags & IFF_RUNNING) {
	hcf_disable(&sc->ifb, 0);
	ifp->if_flags &= ~IFF_RUNNING;
      }
    }

#if NBPFILTER > 0
    /*
     * Promiscuous flag may have changed, so reprogram the multicast filter.
     */
    iwl_setmulti(ifp);
#endif
    break;

  case SIOCADDMULTI:
  case SIOCDELMULTI:
    /*
     * Multicast list has changed; set the hardware filter
     * accordingly.
     */
    iwl_setmulti(ifp);
    error = 0;
    break;

  case SIOCGIWLSTATS:
    ifr = (struct ifreq *) data;
    hcf_action(&sc->ifb, HCF_ACT_TALLIES);
    error = copyout(&sc->ifb.IFB_NIC_Tallies, ifr->ifr_data,
		    sizeof(struct iwl_hermes_tallies));
    break;

  case SIOCGIWLSSID:
    ifr = (struct ifreq *) data;
    iwl_hw_getssid(&sc->ifb, tnam, sizeof(tnam));
    error = copyout(tnam, ifr->ifr_data, strlen(tnam)+1);
    break;

  case SIOCSIWLSSID:
    if (!(error = suser(p->p_ucred, &p->p_acflag))) {
      ifr = (struct ifreq *) data;
      if (!(error = copyinstr(ifr->ifr_data, tnam, sizeof(tnam), &done))) {
	if ((error = iwl_hw_setssid(&sc->ifb, tnam)) ||
	    (error = hcf_enable(&sc->ifb, 0)))
	  error = EIO;
      }
    }
    break;

  case SIOCGIWLPTYPE:
    error = iwl_hw_getporttype(&sc->ifb);
    ifr = (struct ifreq *) data;
    if (error == -1)
      error = EIO;
    else
      error = copyout(&error, ifr->ifr_data, sizeof(int));
    break;

  case SIOCSIWLPTYPE:
    if (!(error = suser(p->p_ucred, &p->p_acflag))) {
      ifr = (struct ifreq *) data;
      if (!(error = copyin(ifr->ifr_data, &i, sizeof(int)))) {
	if (i != 1 && i != 3)
	  error = EINVAL;
	else {
	  if ((error = iwl_hw_setporttype(&sc->ifb, i)) ||
	      (error = hcf_enable(&sc->ifb, 0)))
	    error = EIO;
	}
      }
    }
    break;

  case SIOCGIWLCHAN:
    error = iwl_hw_getchannel(&sc->ifb);
    ifr = (struct ifreq *) data;
    if (error == -1)
      error = EIO;
    else
      error = copyout(&error, ifr->ifr_data, sizeof(int));
    break;

  case SIOCSIWLCHAN:
    if (!(error = suser(p->p_ucred, &p->p_acflag))) {
      ifr = (struct ifreq *) data;
      if (!(error = copyin(ifr->ifr_data, &i, sizeof(int)))) {
	if (i < 1 || i > 14)
	  error = EINVAL;
	else {
	  if ((error = iwl_hw_setchannel(&sc->ifb, i)) ||
	      (error = hcf_enable(&sc->ifb, 0)))
	    error = EIO;
	}
      }
    }
    break;

  case SIOCGIWLPWR:
    error = iwl_hw_getpower(&sc->ifb);
    ifr = (struct ifreq *) data;
    if (error == -1)
      error = EIO;
    else
      error = copyout(&error, ifr->ifr_data, sizeof(int));
    break;

  case SIOCSIWLPWR:
    if (!(error = suser(p->p_ucred, &p->p_acflag))) {
      ifr = (struct ifreq *) data;
      if (!(error = copyin(ifr->ifr_data, &i, sizeof(int)))) {
	if (i < 0 || i > 1)
	  error = EINVAL;
	else {
	  if ((error = iwl_hw_setpower(&sc->ifb, i)) ||
	      (error = hcf_enable(&sc->ifb, 0)))
	    error = EIO;
	}
      }
    }
    break;

  case SIOCGIWLRTXS:
    error = iwl_hw_getrtxs(&sc->ifb);
    ifr = (struct ifreq *) data;
    if (error == -1)
      error = EIO;
    else
      error = copyout(&error, ifr->ifr_data, sizeof(int));
    break;

  case SIOCSIWLRTXS:
    error = EINVAL;
    break;

  case SIOCGIWLRTXL:
    error = iwl_hw_getrtxl(&sc->ifb);
    ifr = (struct ifreq *) data;
    if (error == -1)
      error = EIO;
    else
      error = copyout(&error, ifr->ifr_data, sizeof(int));
    break;

  case SIOCSIWLRTXL:
    error = EINVAL;
    break;

  default:
    error = EINVAL;
  }
  (void) splx(s);
#ifdef IWL_DEBUG
  printf("<- iwl_ioctl()\n");
#endif
  return (error);
}

#ifdef 0
struct iw_statistics *wvlan_get_wireless_stats (struct device *dev)
{
	struct net_local *local = (struct net_local *) dev->priv;
	CFG_COMMS_QUALITY_STRCT ltv;
	int rc;

	DEBUG(DEBUG_CALLTRACE, "-> wvlan_get_wireless_stats(%s)\n", dev->name);

	// TODO: Status
	local->wstats.status = 0;

	// Link quality
	ltv.len = 4;
	ltv.typ = CFG_COMMS_QUALITY;
	rc = hcf_get_info(&local->ifb, (LTVP) &ltv);
	DEBUG(DEBUG_NOISY, "%s: hcf_get_info(CFG_COMMS_QUALITY) returned 0x%x\n", dev_info, rc);
	local->wstats.qual.qual = ltv.coms_qual;
	local->wstats.qual.level = ltv.signal_lvl;
	local->wstats.qual.noise = ltv.noise_lvl;
	local->wstats.qual.updated = 1;

	// Packets discarded in the wireless adapter due to wireless specific problems
	local->wstats.discard.nwid = 0;
	local->wstats.discard.code = local->ifb.IFB_NIC_Tallies.RxWEPUndecryptable;
	local->wstats.discard.misc = local->ifb.IFB_NIC_Tallies.RxFCSErrors +
					local->ifb.IFB_NIC_Tallies.RxDiscards_NoBuffer +
					local->ifb.IFB_NIC_Tallies.TxDiscardsWrongSA;

	DEBUG(DEBUG_CALLTRACE, "<- wvlan_get_wireless_stats()\n");
	return (&local->wstats);
}
#endif

static char iwl_pad[68] = { 0 };

static void iwl_start (struct ifnet *ifp)
{
  struct iwl_softc *sc = ifp->if_softc;
  int rc, s, len;
  struct mbuf *m, *m0;
  caddr_t p;

#ifdef IWL_DEBUG
  printf("-> iwl_start()\n");
#endif

  if (sc->gone)
    return;

  s = splimp();
  /* it appears that the bizarro OS-independent layer supplied by Lucent
     utterly refuses to tell us how much buffer space is available on the NIC,
     or even whether our writes to the NIC complete entirely! So we've no
     choice but to write a single packet to the NIC at a time, and not write
     to it again until the prior write completes. This bites: the NIC can't
     stream out packets, as there'll always be some delay before the i/f
     output code wakes up to try transmitting again. */

  if (ifp->if_flags & IFF_OACTIVE) {
    (void)splx(s);
    return;
  }

  IF_DEQUEUE(&ifp->if_snd, m);
  if (m == 0) {
    /* no packets to send. */
    (void)splx(s);
    return;
  }

  /* XXX: only one packet in Tx buffer at a time! ugh. */
  ifp->if_flags |= IFF_OACTIVE;

  /* Send packet */
  m0 = m;
  for (len = 0; m != 0; m = m->m_next)
    len += m->m_len;

  /* 802.11 WaveLAN base stations use LLC SNAP headers on all Ethernet packets.
     So do hosts using Lucent-supplied drivers (NT, Win95). Encapsulate
     FreeBSD's Ethernet-II frames. */
  if ((m0 = m_pullup(m0, 14)) == 0) {
    /* m_pullup failed; not enough data in mbuf chain */
    printf("iwl%d: tx pkt too short: %d bytes!\n", ifp->if_unit, len);
    ifp->if_flags &= ~IFF_OACTIVE;
    (void) splx(s);
    return;
  }
  p = mtod(m0, caddr_t);

  hcf_put_data(&sc->ifb, p, 12, 0);
  len += sizeof(snap_header);
  snap_header[1] = (max(len,68)-0x0e) & 0xff;
  snap_header[0] = (char)((max(len,68)-0x0e) >> 8);
  hcf_put_data(&sc->ifb, snap_header, sizeof(snap_header), 0);
  hcf_put_data(&sc->ifb, p+12, m0->m_len-12, 0);
  m = m0->m_next;

  for ( ; m != 0; m = m->m_next)
    hcf_put_data(&sc->ifb, mtod(m, caddr_t), m->m_len, 0);

  if (len < 68)
    hcf_put_data(&sc->ifb, iwl_pad, 68 - len, 0);

  /* Send packet */
  rc = hcf_send(&sc->ifb, 0);

  /* Re-enable interrupts */
  (void) splx(s);

#ifdef IWL_DEBUG
  printf("iwl%d: Sending 0x%x octets\n", ifp->if_unit, len);
  printf("iwl%d: hcf_send() returned 0x%x\n", ifp->if_unit, rc);
#endif

#if NBPFILTER > 0
  if (ifp->if_bpf) {
    bpf_mtap(ifp, m0);
  }
#endif

  m_freem(m0);
  ifp->if_timer = 2;
#ifdef IWL_DEBUG
  printf("<- iwl_start()\n");
#endif
  return;
}
