/*

  Copyright (C) 1999 President and Fellows of Harvard College

  All rights reserved.

  NOTICE: This software is provided "as is", without any warranty,
  including any implied warranty for merchantability or fitness for a
  particular purpose.  Under no circumstances shall Harvard University
  or its faculty, staff, students or agents be liable for any use of,
  misuse of, or inability to use this software, including incidental
  and consequential damages.

  License is hereby given to use, modify, and redistribute this
  software, in whole or in part, for any commercial or non-commercial
  purpose, provided that the user agrees to the terms of this
  copyright notice, including disclaimer of warranty, and provided
  that this copyright notice, including disclaimer of warranty, is
  preserved in the source code and documentation of anything derived
  from this software.  Any redistributor of this software or anything
  derived from this software assumes responsibility for ensuring that
  any parties to whom wuch a redistribution is made are fully aware of
  the terms of this license and disclaimer.

  Ported from Linux and extended by:
    Brad Karp, Harvard University EECS, April, 1999
    karp@eecs.harvard.edu

  The above copyright and licensing notice applies to the
  *modifications* made at Harvard to port and extend the Linux driver,
  whose original GPL license appears below. Note that the GPL license
  is also in force on this FreeBSD port of the driver.

*/

/* This file is part of the Hardware Control Functions Light
   (HCF-light) library to control the Lucent Technologies WaveLAN/IEEE
   Network I/F Card.  The HCF is the implementation of the Wireless
   Connection I/F (WCI).
   
   The HCF-light files are a subset of the HCF files. The complete set
   offers a number of additional facilities, e.g. firmware download,
   Etherner-II encapsulation, additional diagnostic facilities, ASSERT
   logic to support debugging, 802.11 support, Configuration
   Management.  This complete set is explicitely not in the Public
   Domain but can be made available under certain restriction. (see
   the pointer below for support)
   
   The HCF-light files are free software; you can redistribute it
   and/or modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2 of
   the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
   
   At the time of this writing, you can request for support at:
   betasupport@wavelan.com
   
   Documentation is expected to be available in the week of 8 Februari 1999
*/


#ifndef IWL_HCFCFG_H                                                     
#define IWL_HCFCFG_H 1

/****************************************************************************
*
* FILE	 : hcfcfg.tpl / hcfcfg.h **************************** 2.0 **********
*
* DATE   : $Date:   02 Feb 1999 14:32:32  $   $Revision:   1.0  $
*
* AUTHOR : Nico Valster
*
* DESC   : HCF Customization Macros
*
*****************************************************************************
* COPYRIGHT (c) 1994, 1995		 by AT&T.	 		
*		All Rights Reserved.
* COPYRIGHT (c) 1996, 1997, 1998 by Lucent Technologies.
*	All Rights Reserved.
*
*****************************************************************************/

#include <machine/endian.h>

/*
 * hcfcfg.tpl list all #defines which must be specified to:
 *      I:	adjust the HCF functions defined in HCF.CPP to the
 *              characteristics of a specific environment
 * 		o maximum sizes for messages and notification frames,
 *                persistent configuration storage
 * 		o Endianess
 *
 *	II:	Compiler specific macros
 * 		o port I/O macros
 * 		o type definitions
 *
 *      III:	Environment specific ASSERT macro
 *
 *   IV: Compiler specific 
 *
 *    V: ;? specific 
 *
 *
 * By copying HCFCFG.TPL to HCFCFG.H and -if needed- modifying the #defines
 * the WCI functionality can be tailored
 * T O   D O :  A D D   A   R E C I P E   T O   D O   T H I S
 *
 */

/*	Endianess
 *	Default: HCF_LITTLE_ENDIAN
 *	Little Endian (a.k.a. Intel), least significant byte first
 *	Big Endian (a.k.a. Motorola), most significant byte first
 *
 * If neither HCF_LITTLE_ENDIAN nor HCF_BIG_ENDIAN, the defintion of the following macros must be supplied
 * by the MSF programmer:
 *  o CNV_LITTLE_TO_INT(w)			interprets the 16-bits input value as Little Endian, returns an hcf_16
 * 	o CNV_BIG_TO_INT(w)				interprets the 16-bits input value as Big Endian, returns an hcf_16
 * 	o CNV_INT_TO_BIG_NP(addr)		converts in place the 16-bit value addressed by a near pointer from hcf_16
 * 									to Big Endian
 * 	o CNV_LITTLE_TO_INT_NP(addr)	converts in place the 16-bit value addressed by a near pointer from 
 *									Little endian to hcf_16
 *
 * At a number of places in the HCF code, the CNV_INT_TO_BIG_NP macro is used. While it does have the desired 
 * effect on all platforms, it's naming is misleading, so revisit all places where these CNV macros are used
 * to assure the right name is used at the right place.
 * Hint: introduce CNV_HOST_TO_NETWORK names if appropriate
 *
 */
 
/*	Alignment
 *	Some platforms can access words on odd boundaries (with possibly an performance impact), at other
 *	platforms such an access may result in a memory access violation.
 *	It is assumed that everywhere where the HCF casts a char pointer into a word pointer, the 
 *	alignment criteria are met. This put some restrictions on the MSF, which are assumed to be 
 *	"automatically" fullfilled at the applicable platforms
 *	To assert this assumption, the macro HCF_ALIGN can be defined. The default vaslue is 0, meaning no
 *	alignment, a value of 2 means word alignment, other values are invalid
 */

/*  * * * * * * * * * * * * * * * * * * * * * * II * * * * * * * * * * * * * * * * * * * * * * */



/************************************************************************************************/
/******************  C O M P I L E R   S P E C I F I C   M A C R O S  ***************************/
/************************************************************************************************/
/*************************************************************************************************
*
* The platforms supported by this version are:
*	- Microsoft Visual C 1.5 (16 bits platform)
*	- Microsoft Visual C 2.0 (32 bits platform)
*	- Watcom C/C++ 9.5
*	- SCO UNIX
*
* In this version of hcfiocfg.tpl all macros except the MSVC 1.5 versions are either dependent on
* compiler/environment supplied macros (e.g. _MSC_VER or "def-ed out"
*
* By selecting the appropriate Macro definitions by means of modifying the
* "#ifdef 0/1" lines, the HCF can be adjusted for the I/O chararcteristics of
* a specific compiler
*
* If needed the macros can be modified or replaced with definitions appropriate
* for your personal platform
* If you need to make such changes it is appreciated if you inform Lucent Technologies WCND Utrecht
* That way the changes can become part of the next release of the WCI
*
*
*	The prototypes and functional description of the macros are:
*
*	hcf_8	IN_PORT_BYTE(  hcf_16 port)
*			Reads a byte (8 bits) from the specified port
*
*	hcf_16	IN_PORT_WORD(  hcf_16 port)
*			Reads a word (16 bits) from the specified port
*
*	void	OUT_PORT_BYTE( hcf_16 port, hcf_8 value)
*			Writes a byte (8 bits) to the specified port
*
*	void	OUT_PORT_WORD( hcf_16 port, hcf_16 value)
*			Writes a word (16 bits) to the specified port
*
*	void	IN_PORT_STRING( port, dest, len)
*			Reads len number of words from the specified port to the (FAR) address dest in PC-RAM
*			Note that len specifies the number of words, NOT the number of bytes
*			!!!NOTE, although len specifies the number of words, dest MUST be a char pointer NOTE!!!
*			See also the common notes for IN_PORT_STRING and OUT_PORT_STRING
*
*	void	OUT_PORT_STRING( port, src, len)
*			Writes len number of words from the (FAR) address src in PC-RAM to the specified port
*			Note that len specifies the number of words, NOT the number of bytes.
*			!!!NOTE, although len specifies the number of words, src MUST be a char pointer NOTE!!!
*
*			The peculiar combination of word-length and char pointers for IN_PORT_STRING as well as
*			OUT_PORT_STRING is justified by the assumption that it offers a more optimal algorithm
*
*			Note to the HCF-implementor:
*			Due to the passing of the parameters to compiler specific blabla.........
*			do not use "expressions" as parameters, e.g. don't use "ifbp->IFB_IOBase + HREG_AUX_DATA" but
*			assign this to a temporary variable.
*
*
*  NOTE!!	For convenience of the MSF-programmer, all {IN|OUT}_PORT_{BYTE|WORD|STRING} macros are allowed to 
*			modify their parameters (although some might argue that this would constitute bad coding
*			practice). This has its implications on the HCF, e.g. as a consequence these macros should not
*			be called with parameters which have side effects, e.g auto-increment.
*
*  NOTE!!	in the Micosoft implementation of inline assembly it is O.K. to corrupt all flags except
*			the direction flag and to corrupt all registers except the segment registers and EDI, ESI, 
*			ESP and EBP (or their 16 bits equivalents).
*			Other environments may have other constraints
*
*  NOTE!!	in the Intel environment it is O.K to have a word (as a 16 bits quantity) at a byte boundary, 
*			hence IN_/OUT_PORT_STRING can move words between PC-memory and NIC-memory with as only
*			constraint that the the words are on a word boundary in NIC-memory. This does not hold true
*			for all conceivalble environments, e.g. an Motorola 68xxx does not allow this, in other
*			words whenever there is a move from address in 2*n in one memory type to address 2*m+1 in the
*			other type, the current templates for IN_/OUT_PORT_STRING are unsuitable. Probably the
*			boundary conditions imposed by these type of platforms prevent this case from materializing
*
*************************************************************************************************/

#if defined __FreeBSD__

#define HCF_STA			/* station characteristics */
#define	MSF_COMPONENT_ID COMP_ID_LINUX
#define	MSF_COMPONENT_VAR 1
#define	MSF_COMPONENT_MAJOR_VER 0
#define	MSF_COMPONENT_MINOR_VER 4

#include <sys/types.h>
#include <sys/cdefs.h>
#include <machine/cpufunc.h>

/* #define CNV_LITTLE_TO_INT(x)		/* No endianess conversion needed */

#define IN_PORT_BYTE(port) ((uint8_t)inb( (hcf_io)(port) ))
#define IN_PORT_WORD(port) ((uint16_t)inw( (hcf_io)(port) ))
#define OUT_PORT_BYTE(port, value) (outb( (hcf_io)(port), (uint8_t) (value)))
#define OUT_PORT_WORD(port, value) (outw( (hcf_io)(port), (uint16_t) (value)))

#define IN_PORT_STRING(prt, dst, n) while ( n-- ) { *(uint16_t *)dst = IN_PORT_WORD(prt); dst += 2; }
#define OUT_PORT_STRING(prt, src, n) while ( n-- ) { OUT_PORT_WORD(prt, *(uint16_t *)src) ; src  += 2; }

#endif	/* FreeBSD */

/* * * * * * * * * * * * * * * * * * * * * * *  IV  * * * * * * * * * * * */

/***************************************Compiler specific *****************/

/**************************************************************************/
/* M A C R O S derived of C O M P I L E R   S P E C I F I C   M A C R O S */
/**************************************************************************/
uint8_t foo;

typedef uint8_t *wci_bufp;	/* segmented 16-bits or flat 32-bits pointer
				   to 8 bits unit */
typedef uint16_t *wci_recordp;	/* segmented 16-bits or flat 32-bits pointer
				   to 16 bits unit */

typedef uint8_t *hcf_8fp;	/* segmented 16-bits or flat 32-bits pointer
				   to 8 bits unit */
typedef uint16_t *hcf_16fp;	/* segmented 16-bits or flat 32-bits pointer
				   to 16 bits unit */
typedef uint32_t *hcf_32fp;	/* segmented 16-bits or flat 32-bits pointer
				   to 8 bits unit */

#if defined HCF_STA && defined HCF_AP
#if defined WCITST
#pragma message( "you should also test both in isolation" )
#else
error; define exactly one of these terms;
#endif
#endif
#if ! defined HCF_STA && ! defined HCF_AP
error; define exactly one of these terms;
#endif

/* * * * * * * * * * * * * * * * * * * * * * *  V  * * * * * * * * * * * * */

typedef uint16_t hcf_io;

/* MSF_COMPONENT_ID is used to define the CFG_IDENTITY_STRCT in HCF.C
 * CFG_IDENTITY_STRCT is defined in HCF.C purely based on convenience
 * arguments The HCF can not have the knowledge to determine the
 * ComponentId field of the Identity record (aka as Version Record),
 * therefore the MSF part of the Drivers must supply this value via
 * the System Constant MSF_COMPONENT_ID There is a set of values
 * predefined in MDD.H (format COMP_ID_.....)  */

#if defined MSF_COMPONENT_ID
#define	DUI_COMPAT_VAR MSF_COMPONENT_ID
#define	DUI_COMPAT_BOT 4
#define	DUI_COMPAT_TOP 4
#endif /* MSF_COMPONENT_ID */

/****************************************************************************/
/***************  E N V I R O N M E N T   S P E C I F I C   M A C R O S  ****/
/****************************************************************************/

#if defined HCF_ASSERT
/* the next line may need modification for a specific environment */
extern int HCF_VAR_0; /* Revision 2.0 gives an explanation of the need for
			 HCF_VAR_0 */
#endif

/****************************************************************************/
/******  M S F    S U P P O R T    F U N C T I O N S    P R O T O T Y P E S */
/****************************************************************************/

void msf_assert (wci_bufp file_namep, unsigned int line_number,
		 uint16_t trace, int qual);

/* To increase portability, use unsigned char and unsigned char * when
 * accessing parts of larger types to convert their Endianess */

#if BYTE_ORDER == BIG_ENDIAN
/************************************* B I G   E N D I A N ******************/
#define CNV_LITTLE_TO_INT(w) ( ((uint16_t)(w) & 0x00ff) << 8 |    \
			       ((uint16_t)(w) & 0xff00) >> 8 )
#define CNV_BIG_TO_INT(w) (w)	/* No endianess conversion needed */

#define CNV_INT_TO_BIG_NP(addr)
#define CNV_LITTLE_TO_INT_NP(addr) {                              \
	uint8_t temp;                                             \
	temp = ((uint8_t *)(addr))[0];                            \
	((uint8_t *)(addr))[0] = ((uint8_t *)(addr))[1];          \
	((uint8_t *)(addr))[1] = temp;                            \
}

#endif /* BIG_ENDIAN */

#if BYTE_ORDER == LITTLE_ENDIAN
/************************************* L I T T L E   E N D I A N ************/

#define CNV_LITTLE_TO_INT(w) (w) /* No endianess conversion needed */
#define CNV_BIG_TO_INT(w) ( ((uint16_t)(w) & 0x00ff) << 8 |       \
			    ((uint16_t)(w) & 0xff00) >> 8 )

#define CNV_INT_TO_BIG_NP(addr) {				  \
	uint8_t temp;						  \
	temp = ((uint8_t *)(addr))[0];				  \
	((uint8_t *)(addr))[0] = ((uint8_t *)(addr))[1];	  \
	((uint8_t *)(addr))[1] = temp;				  \
}
#define CNV_LITTLE_TO_INT_NP(addr)

#endif /* LITTLE_ENDIAN */

/* conversion macros which can be expressed in other macros */
#define CNV_INT_TO_LITTLE(w) CNV_LITTLE_TO_INT(w)
#define CNV_INT_TO_BIG(w) CNV_BIG_TO_INT(w)

#endif /* HCFCFG_H */

/******************************************* A L I G N M E N T  *************/
#if defined HCF_ALIGN 
#if HCF_ALIGN != 0 && HCF_ALIGN != 2
	error;
#endif /* HCF_ALIGN != 0 && HCF_ALIGN != 2 */
#else
#define HCF_ALIGN 0		/* default to no alignment */
#endif /* HCF_ALIGN */

/*  * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *	
 *	The routines ips and ops (short for InPutString and OutPutString) are
 *      created to use the compiler to do the type checking. It turned out
 *      that it is too easy to accidentally pass a word pointer to the the
 *      macros IN_PORT_STRING and OUT_PORT_STRING rather than a byte pointer.
 *	The "+2" as some macro implementations use, does not have the intended
 *      effect in those cases.
 *	The HCF_STRICT business can be ignored by MSF programmers.
 *	
 *  * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#if defined HCF_STRICT
void ips(hcf_io prt, wci_bufp dst, int n);
void ops(hcf_io prt, wci_bufp src, int n);
#endif /* HCF_STRICT */

#if !defined HCF_MAX_CONFIG
#define HCF_MAX_CONFIG 256	/* maximum accumulated size in hcf_16 of LTV
				   records used in hcf_put_config */
#endif

#if !defined HCF_MAX_MSG
#define HCF_MAX_MSG 2304	/* WaveLAN Packet Size */
#endif

#if !defined HCF_MAX_NOTIFY		
#define HCF_MAX_NOTIFY 6	/* maximum size in bytes of "real" data in
				   Notify command */
#endif
